# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from typing import Any, Dict, List, Literal, Mapping, Optional, Set, TypeVar, Union

from glide.protobuf.command_request_pb2 import CommandRequest
from glide.protobuf.connection_request_pb2 import ConnectionRequest
from glide.routes import ByAddressRoute, RandomNode, SlotIdRoute, SlotKeyRoute

OK: str = "OK"
DEFAULT_READ_BYTES_SIZE: int = pow(2, 16)
# Typing
T = TypeVar("T")
TOK = Literal["OK"]
TResult = Union[
    TOK,
    str,
    List[List[str]],
    int,
    None,
    Dict[str, T],
    Mapping[str, "TResult"],
    float,
    Set[T],
    List[T],
    bytes,
    Dict[bytes, "TResult"],
    Mapping[bytes, "TResult"],
]
TRequest = Union[CommandRequest, ConnectionRequest]
# When routing to a single node, response will be T
# Otherwise, response will be : {Address : response , ... } with type of Dict[str, T].
TClusterResponse = Union[T, Dict[bytes, T]]
TSingleNodeRoute = Union[RandomNode, SlotKeyRoute, SlotIdRoute, ByAddressRoute]
# When specifying legacy path (path doesn't start with `$`), response will be T
# Otherwise, (when specifying JSONPath), response will be List[Optional[T]].
#
# TJsonResponse is designed to handle scenarios where some paths may not contain valid values, especially with JSONPath
# targeting multiple paths.
# In such cases, the response may include None values, represented as `Optional[T]` in the list.
# This type provides flexibility for commands where a subset of the paths may return None.
#
# For more information, see: https://redis.io/docs/data-types/json/path/ .
TJsonResponse = Union[T, List[Optional[T]]]

# When specifying legacy path (path doesn't start with `$`), response will be T
# Otherwise, (when specifying JSONPath), response will be List[T].
# This type represents the response format for commands that apply to every path and every type in a JSON document.
# It covers both singular and multiple paths, ensuring that the command returns valid results for each matched path
# without None values.
#
# TJsonUniversalResponse is considered "universal" because it applies to every matched path and
# guarantees valid, non-null results across all paths, covering both singular and multiple paths.
# This type is used for commands that return results from all matched paths, ensuring that each
# path contains meaningful values without None entries (unless it's part of the commands response).
# It is typically used in scenarios where each target is expected to yield a valid response. For commands that are valid
# for all target types.
#
# For more information, see: https://redis.io/docs/data-types/json/path/ .
TJsonUniversalResponse = Union[T, List[T]]
TEncodable = Union[str, bytes]
TFunctionListResponse = List[
    Mapping[
        bytes,
        Union[bytes, List[Mapping[bytes, Union[bytes, Set[bytes]]]]],
    ]
]
# Response for function stats command on a single node.
# The response holds a map with 2 keys: Current running function / script and information about it, and the engines and
# the information about it.
TFunctionStatsSingleNodeResponse = Mapping[
    bytes,
    Union[
        None,
        Mapping[
            bytes,
            Union[Mapping[bytes, Mapping[bytes, int]], bytes, int, List[bytes]],
        ],
    ],
]
# Full response for function stats command across multiple nodes.
# It maps node address to the per-node response.
TFunctionStatsFullResponse = Mapping[
    bytes,
    TFunctionStatsSingleNodeResponse,
]


TXInfoStreamResponse = Mapping[
    bytes, Union[bytes, int, Mapping[bytes, Optional[List[List[bytes]]]]]
]
TXInfoStreamFullResponse = Mapping[
    bytes,
    Union[
        bytes,
        int,
        Mapping[bytes, List[List[bytes]]],
        List[Mapping[bytes, Union[bytes, int, List[List[Union[bytes, int]]]]]],
    ],
]

FtInfoResponse = Mapping[
    TEncodable,
    Union[
        TEncodable,
        int,
        List[TEncodable],
        List[
            Mapping[
                TEncodable,
                Union[TEncodable, Mapping[TEncodable, Union[TEncodable, int]]],
            ]
        ],
    ],
]

FtSearchResponse = List[
    Union[int, Mapping[TEncodable, Mapping[TEncodable, TEncodable]]]
]

FtAggregateResponse = List[Mapping[TEncodable, Any]]

FtProfileResponse = List[
    Union[FtSearchResponse, FtAggregateResponse, Mapping[str, int]]
]



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import asyncio

import pytest
from glide.config import ProtocolVersion
from glide.constants import OK
from glide.exceptions import RequestError
from glide.glide_client import TGlideClient

from tests.conftest import (
    NEW_PASSWORD,
    USERNAME,
    WRONG_PASSWORD,
    auth_client,
    config_set_new_password,
    kill_connections,
)
from tests.utils.utils import (
    delete_acl_username_and_password,
    set_new_acl_username_with_password,
)


@pytest.mark.asyncio
class TestAuthCommands:
    """Test cases for password authentication and management"""

    @pytest.fixture(autouse=True, scope="function")
    async def cleanup(self, request, management_client: TGlideClient):
        """
        Ensure password is reset for default user and USERNAME user is deleted after each test, regardless of test outcome.
        This fixture runs after each test.
        """
        yield
        try:
            # reset password for default user
            await auth_client(management_client, NEW_PASSWORD)
            await config_set_new_password(management_client, "")
            await management_client.update_connection_password(None)

            # delete USERNAME user
            await delete_acl_username_and_password(management_client, USERNAME)
        except RequestError:
            pass

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing the connection password without immediate re-authentication.
        Verifies that:
        1. The client can update its internal password
        2. The client remains connected with current auth
        3. The client can reconnect using the new password after server password change
        This test is only for cluster mode, as standalone mode does not have a connection available handler
        """
        result = await glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=False
        )
        assert result == OK
        # Verify that the client is still authenticated
        assert await glide_client.set("test_key", "test_value") == OK
        value = await glide_client.get("test_key")
        assert value == b"test_value"
        await config_set_new_password(glide_client, NEW_PASSWORD)
        await kill_connections(management_client)
        # Add a short delay to allow the server to apply the new password
        # without this delay, command may or may not time out while the client reconnect
        # ending up with a flaky test
        await asyncio.sleep(2)
        # Verify that the client is able to reconnect with the new password,
        value = await glide_client.get("test_key")
        assert value == b"test_value"
        await kill_connections(management_client)
        await asyncio.sleep(2)
        # Verify that the client is able to immediateAuth with the new password after client is killed
        result = await glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=True
        )
        assert result == OK
        # Verify that the client is still authenticated
        assert await glide_client.set("test_key", "test_value") == OK

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_connection_lost_before_password_update(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test changing server password when connection is lost before password update.
        Verifies that the client will not be able to reach the inner core and return an error
        on immediate re-authentication, but will succeed with non-immediate re-auth
        """
        await glide_client.set("test_key", "test_value")
        await config_set_new_password(glide_client, NEW_PASSWORD)
        await kill_connections(management_client)
        await asyncio.sleep(2)
        result = await glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=False
        )
        assert result == OK
        with pytest.raises(RequestError):
            await glide_client.update_connection_password(
                NEW_PASSWORD, immediate_auth=True
            )

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_no_server_auth(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test that immediate re-authentication fails when no server password is set.
        This verifies proper error handling when trying to re-authenticate with a
        password when the server has no password set.
        """
        with pytest.raises(RequestError):
            await glide_client.update_connection_password(
                WRONG_PASSWORD, immediate_auth=True
            )

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_long(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with a long password string.
        Verifies that the client can handle long passwords (1000 characters).
        """
        long_password = "p" * 1000
        result = await glide_client.update_connection_password(
            long_password, immediate_auth=False
        )
        assert result == OK

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_replace_password_immediate_auth_wrong_password(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test that re-authentication fails when using wrong password.
        Verifies proper error handling when immediate re-authentication is attempted
        with a password that doesn't match the server's password.
        """
        await config_set_new_password(glide_client, NEW_PASSWORD)
        with pytest.raises(RequestError):
            await glide_client.update_connection_password(
                WRONG_PASSWORD, immediate_auth=True
            )

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_with_immediate_auth(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with immediate re-authentication.
        Verifies that:
        1. The client can update its password and re-authenticate immediately
        2. The client remains operational after re-authentication
        """
        await config_set_new_password(glide_client, NEW_PASSWORD)
        result = await glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=True
        )
        assert result == OK
        # Verify that the client is still authenticated
        assert await glide_client.set("test_key", "test_value") == OK
        value = await glide_client.get("test_key")
        assert value == b"test_value"

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_auth_non_valid_pass(
        self, glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with immediate re-authentication using a non-valid password.
        Verifies that immediate re-authentication fails when the password is not valid.
        """
        with pytest.raises(RequestError):
            await glide_client.update_connection_password(None, immediate_auth=True)
        with pytest.raises(RequestError):
            await glide_client.update_connection_password("", immediate_auth=True)

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_with_acl_user(
        self, acl_glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing the connection password for an ACL user without immediate re-authentication.
        and not the default one.
        Verifies that:
        1. The client can update its internal password for the ACL user
        2. The client remains connected with current auth
        3. The client can reconnect using the new password after server password change (which is simulated by
        deleting and reseting the user with a new password, which kills the connection).
        """

        # Create a new ACL user and authenticate the client as the new user
        await acl_glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=False
        )

        # Verify that the client is authenticated
        assert await acl_glide_client.set("test_key", "test_value") == OK
        value = await acl_glide_client.get("test_key")
        assert value == b"test_value"

        # Delete the username and reset it with new password (equivalent to config_set new password)
        assert await delete_acl_username_and_password(management_client, USERNAME) == 1
        await set_new_acl_username_with_password(
            management_client, USERNAME, NEW_PASSWORD
        )

        # Sleep to allow enough time for reconnecting
        await asyncio.sleep(2)

        # The client should now reconnect with the new password automatically
        # Verify that the client is still able to perform operations
        value = await acl_glide_client.get("test_key")
        assert value == b"test_value"

        await acl_glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=True
        )

        assert await acl_glide_client.set("new_key", "new_value") == OK
        value = await acl_glide_client.get("new_key")
        assert value == b"new_value"

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_reconnection_with_immediate_auth_with_acl_user(
        self, acl_glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with immediate re-authentication.
        Verifies that:
        1. Upon disconnection (which is caused by the user deletion), the client succeeds in re-authentication
            with the correct password.
        2. The client remains operational after re-authentication
        This test is relevant only for cluster mode - in standalone, reconnection will fail and new requests for
        the server won't be served.
        """
        assert await delete_acl_username_and_password(management_client, USERNAME) == 1
        await set_new_acl_username_with_password(
            management_client, USERNAME, NEW_PASSWORD
        )

        # Sleep to allow enough time for reconnecting
        await asyncio.sleep(2)

        result = await acl_glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=True
        )
        assert result == OK

        # Verify client is authenticated
        assert await acl_glide_client.set("test_key", "test_value") == OK
        value = await acl_glide_client.get("test_key")
        assert value == b"test_value"

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_connection_lost_before_password_update_acl_user(
        self, acl_glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with immediate re-authentication.
        Verifies that:
        1. Upon disconnection (which is caused by the user deletion), the client succeeds in updating the password
        with non-immediate auth (this is an internal operation not requiring a server connection).
        2. Trying to connect with immediate authentication fails due to reconnection attempts with the previous password.
        This test is relevant only for standalone - in standalone, reconnection will fail and new requests for
        the server won't be served.
        """
        assert await delete_acl_username_and_password(management_client, USERNAME) == 1
        await set_new_acl_username_with_password(
            management_client, USERNAME, NEW_PASSWORD
        )

        # ensure client disconnection
        await asyncio.sleep(2)

        with pytest.raises(RequestError):
            await acl_glide_client.update_connection_password(
                NEW_PASSWORD, immediate_auth=True
            )

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_replace_password_immediateAuth_acl_user(
        self, acl_glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Tests adding a new password to the user, verifies that the client succeeds in immediate authentication with it.
        """
        await set_new_acl_username_with_password(
            management_client, USERNAME, NEW_PASSWORD
        )

        result = await acl_glide_client.update_connection_password(
            NEW_PASSWORD, immediate_auth=True
        )

        assert result == OK

        assert await acl_glide_client.set("test_key", "test_value") == OK
        value = await acl_glide_client.get("test_key")
        assert value == b"test_value"

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_update_connection_password_auth_non_valid_pass_acl_user(
        self, acl_glide_client: TGlideClient, management_client: TGlideClient
    ):
        """
        Test replacing connection password with immediate re-authentication using a non-valid password.
        Verifies that immediate re-authentication fails when the password is not valid.
        """
        with pytest.raises(RequestError):
            await acl_glide_client.update_connection_password(None, immediate_auth=True)
        with pytest.raises(RequestError):
            await acl_glide_client.update_connection_password("", immediate_auth=True)



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0
from enum import Enum
from typing import List, Optional, Union, cast

from glide.async_commands.server_modules.ft_options.ft_aggregate_options import (
    FtAggregateOptions,
)
from glide.async_commands.server_modules.ft_options.ft_constants import (
    FtProfileKeywords,
)
from glide.async_commands.server_modules.ft_options.ft_search_options import (
    FtSearchOptions,
)
from glide.constants import TEncodable


class QueryType(Enum):
    """
    This class represents the query type being profiled.
    """

    AGGREGATE = "AGGREGATE"
    """
    If the query being profiled is for the FT.AGGREGATE command.
    """
    SEARCH = "SEARCH"
    """
    If the query being profiled is for the FT.SEARCH command.
    """


class FtProfileOptions:
    """
    This class represents the arguments/options for the FT.PROFILE command.
    """

    def __init__(
        self,
        query: TEncodable,
        query_type: QueryType,
        query_options: Optional[Union[FtSearchOptions, FtAggregateOptions]] = None,
        limited: Optional[bool] = False,
    ):
        """
        Initialize a new FtProfileOptions instance.

        Args:
            query (TEncodable): The query that is being profiled. This is the query argument from the
                FT.AGGREGATE/FT.SEARCH command.
            query_type (Optional[QueryType]): The type of query to be profiled.
            query_options (Optional[Union[FtSearchOptions, FtAggregateOptions]]): The arguments/options for the
                FT.AGGREGATE/FT.SEARCH command being profiled.
            limited (Optional[bool]): To provide some brief version of the output, otherwise a full verbose output is provided.
        """
        self.query = query
        self.query_type = query_type
        self.query_options = query_options
        self.limited = limited

    @classmethod
    def from_query_options(
        cls,
        query: TEncodable,
        query_options: Union[FtSearchOptions, FtAggregateOptions],
        limited: Optional[bool] = False,
    ):
        """
        A class method to create FtProfileOptions with FT.SEARCH/FT.AGGREGATE options.

        Args:
            query (TEncodable): The query that is being profiled. This is the query argument from the
                FT.AGGREGATE/FT.SEARCH command.
            query_options (Optional[Union[FtSearchOptions, FtAggregateOptions]]): The arguments/options for the
                FT.AGGREGATE/FT.SEARCH command being profiled.
            limited (Optional[bool]): To provide some brief version of the output, otherwise a full verbose output is provided.
        """
        query_type: QueryType = QueryType.SEARCH
        if isinstance(query_options, FtAggregateOptions):
            query_type = QueryType.AGGREGATE
        return cls(query, query_type, query_options, limited)

    @classmethod
    def from_query_type(
        cls, query: TEncodable, query_type: QueryType, limited: Optional[bool] = False
    ):
        """
        A class method to create FtProfileOptions with QueryType.

        Args:
            query (TEncodable): The query that is being profiled. This is the query argument from the
                FT.AGGREGATE/FT.SEARCH command.
            query_type (QueryType): The type of query to be profiled.
            limited (Optional[bool]): To provide some brief version of the output, otherwise a full verbose output is provided.
        """
        return cls(query, query_type, None, limited)

    def to_args(self) -> List[TEncodable]:
        """
        Get the remaining arguments for the FT.PROFILE command.

        Returns:
            List[TEncodable]: A list of remaining arguments for the FT.PROFILE command.
        """
        args: List[TEncodable] = [self.query_type.value]
        if self.limited:
            args.append(FtProfileKeywords.LIMITED)
        args.extend([FtProfileKeywords.QUERY, self.query])
        if self.query_options:
            if isinstance(self.query_options, FtAggregateOptions):
                args.extend(cast(FtAggregateOptions, self.query_options).to_args())
            else:
                args.extend(cast(FtSearchOptions, self.query_options).to_args())
        return args



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from glide.async_commands.core import CoreCommands
from glide.exceptions import ConfigurationError
from glide.protobuf.connection_request_pb2 import ConnectionRequest
from glide.protobuf.connection_request_pb2 import ProtocolVersion as SentProtocolVersion
from glide.protobuf.connection_request_pb2 import ReadFrom as ProtobufReadFrom
from glide.protobuf.connection_request_pb2 import TlsMode


class NodeAddress:
    """
    Represents the address and port of a node in the cluster.

    Attributes:
        host (str, optional): The server host. Defaults to "localhost".
        port (int, optional): The server port. Defaults to 6379.
    """

    def __init__(self, host: str = "localhost", port: int = 6379):
        self.host = host
        self.port = port


class ReadFrom(Enum):
    """
    Represents the client's read from strategy.
    """

    PRIMARY = ProtobufReadFrom.Primary
    """
    Always get from primary, in order to get the freshest data.
    """
    PREFER_REPLICA = ProtobufReadFrom.PreferReplica
    """
    Spread the requests between all replicas in a round robin manner.
    If no replica is available, route the requests to the primary.
    """
    AZ_AFFINITY = ProtobufReadFrom.AZAffinity
    """
    Spread the read requests between replicas in the same client's AZ (Aviliablity zone) in a round robin manner,
    falling back to other replicas or the primary if needed
    """
    AZ_AFFINITY_REPLICAS_AND_PRIMARY = ProtobufReadFrom.AZAffinityReplicasAndPrimary
    """
    Spread the read requests among nodes within the client's Availability Zone (AZ) in a round robin manner,
    prioritizing local replicas, then the local primary, and falling back to any replica or the primary if needed.
    """


class ProtocolVersion(Enum):
    """
    Represents the communication protocol with the server.
    """

    RESP2 = SentProtocolVersion.RESP2
    """
    Communicate using RESP2.
    """
    RESP3 = SentProtocolVersion.RESP3
    """
    Communicate using RESP3.
    """


class BackoffStrategy:
    """
    Represents the strategy used to determine how and when to reconnect, in case of connection failures.
    The time between attempts grows exponentially, to the formula rand(0 .. factor * (exponentBase ^ N)), where N
    is the number of failed attempts.
    Once the maximum value is reached, that will remain the time between retry attempts until a reconnect attempt is
    successful.
    The client will attempt to reconnect indefinitely.

    Attributes:
        num_of_retries (int): Number of retry attempts that the client should perform when disconnected from the server,
            where the time between retries increases. Once the retries have reached the maximum value, the time between
            retries will remain constant until a reconnect attempt is succesful.
        factor (int): The multiplier that will be applied to the waiting time between each retry.
        exponent_base (int): The exponent base configured for the strategy.
        jitter_percent (Optional[int]): The Jitter percent on the calculated duration. If not set, a default value will be used.
    """

    def __init__(
        self,
        num_of_retries: int,
        factor: int,
        exponent_base: int,
        jitter_percent: Optional[int] = None,
    ):
        self.num_of_retries = num_of_retries
        self.factor = factor
        self.exponent_base = exponent_base
        self.jitter_percent = jitter_percent


class ServerCredentials:
    """
    Represents the credentials for connecting to a server.

    Attributes:
        password (str): The password that will be used for authenticating connections to the servers.
        username (Optional[str]): The username that will be used for authenticating connections to the servers.
            If not supplied, "default" will be used.
    """

    def __init__(
        self,
        password: str,
        username: Optional[str] = None,
    ):
        self.password = password
        self.username = username


class PeriodicChecksManualInterval:
    """
    Represents a manually configured interval for periodic checks.

    Attributes:
        duration_in_sec (int): The duration in seconds for the interval between periodic checks.
    """

    def __init__(self, duration_in_sec: int) -> None:
        self.duration_in_sec = duration_in_sec


class PeriodicChecksStatus(Enum):
    """
    Represents the cluster's periodic checks status.
    To configure specific interval, see PeriodicChecksManualInterval.
    """

    ENABLED_DEFAULT_CONFIGS = 0
    """
    Enables the periodic checks with the default configurations.
    """
    DISABLED = 1
    """
    Disables the periodic checks.
    """


class AdvancedBaseClientConfiguration:
    """
    Represents the advanced configuration settings for a base Glide client.

    Attributes:
        connection_timeout (Optional[int]): The duration in milliseconds to wait for a TCP/TLS connection to complete.
            This applies both during initial client creation and any reconnections that may occur during request processing.
            **Note**: A high connection timeout may lead to prolonged blocking of the entire command pipeline.
            If not explicitly set, a default value of 250 milliseconds will be used.
    """

    def __init__(self, connection_timeout: Optional[int] = None):
        self.connection_timeout = connection_timeout

    def _create_a_protobuf_conn_request(
        self, request: ConnectionRequest
    ) -> ConnectionRequest:
        if self.connection_timeout:
            request.connection_timeout = self.connection_timeout
        return request


class BaseClientConfiguration:
    """
    Represents the configuration settings for a Glide client.

    Attributes:
        addresses (List[NodeAddress]): DNS Addresses and ports of known nodes in the cluster.
            If the server is in cluster mode the list can be partial, as the client will attempt to map out
            the cluster and find all nodes.
            If the server is in standalone mode, only nodes whose addresses were provided will be used by the
            client.
            For example::

                [
                    {address:sample-address-0001.use1.cache.amazonaws.com, port:6379},
                    {address: sample-address-0002.use2.cache.amazonaws.com, port:6379}
                ].

        use_tls (bool): True if communication with the cluster should use Transport Level Security.
            Should match the TLS configuration of the server/cluster, otherwise the connection attempt will fail
        credentials (ServerCredentials): Credentials for authentication process.
            If none are set, the client will not authenticate itself with the server.
        read_from (ReadFrom): If not set, `PRIMARY` will be used.
        request_timeout (Optional[int]): The duration in milliseconds that the client should wait for a request to
            complete.
            This duration encompasses sending the request, awaiting for a response from the server, and any required
            reconnections or retries.
            If the specified timeout is exceeded for a pending request, it will result in a timeout error. If not
            explicitly set, a default value of 250 milliseconds will be used.
        reconnect_strategy (Optional[BackoffStrategy]): Strategy used to determine how and when to reconnect, in case of
            connection failures.
            If not set, a default backoff strategy will be used.
        client_name (Optional[str]): Client name to be used for the client. Will be used with CLIENT SETNAME command
            during connection establishment.
        protocol (ProtocolVersion): Serialization protocol to be used. If not set, `RESP3` will be used.
        inflight_requests_limit (Optional[int]): The maximum number of concurrent requests allowed to be in-flight
            (sent but not yet completed).
            This limit is used to control the memory usage and prevent the client from overwhelming the server or getting
            stuck in case of a queue backlog.
            If not set, a default value will be used.
        client_az (Optional[str]): Availability Zone of the client.
            If ReadFrom strategy is AZAffinity, this setting ensures that readonly commands are directed to replicas
            within the specified AZ if exits.
            If ReadFrom strategy is AZAffinityReplicasAndPrimary, this setting ensures that readonly commands are directed
            to nodes (first replicas then primary) within the specified AZ if they exist.
        advanced_config (Optional[AdvancedBaseClientConfiguration]): Advanced configuration settings for the client.
    """

    def __init__(
        self,
        addresses: List[NodeAddress],
        use_tls: bool = False,
        credentials: Optional[ServerCredentials] = None,
        read_from: ReadFrom = ReadFrom.PRIMARY,
        request_timeout: Optional[int] = None,
        reconnect_strategy: Optional[BackoffStrategy] = None,
        client_name: Optional[str] = None,
        protocol: ProtocolVersion = ProtocolVersion.RESP3,
        inflight_requests_limit: Optional[int] = None,
        client_az: Optional[str] = None,
        advanced_config: Optional[AdvancedBaseClientConfiguration] = None,
    ):
        self.addresses = addresses
        self.use_tls = use_tls
        self.credentials = credentials
        self.read_from = read_from
        self.request_timeout = request_timeout
        self.reconnect_strategy = reconnect_strategy
        self.client_name = client_name
        self.protocol = protocol
        self.inflight_requests_limit = inflight_requests_limit
        self.client_az = client_az
        self.advanced_config = advanced_config

        if read_from == ReadFrom.AZ_AFFINITY and not client_az:
            raise ValueError(
                "client_az must be set when read_from is set to AZ_AFFINITY"
            )

        if read_from == ReadFrom.AZ_AFFINITY_REPLICAS_AND_PRIMARY and not client_az:
            raise ValueError(
                "client_az must be set when read_from is set to AZ_AFFINITY_REPLICAS_AND_PRIMARY"
            )

    def _create_a_protobuf_conn_request(
        self, cluster_mode: bool = False
    ) -> ConnectionRequest:
        """
        Generates a Protobuf ConnectionRequest using the values from this ClientConfiguration.

        Args:
            cluster_mode (bool, optional): The cluster mode of the client. Defaults to False.

        Returns:
            ConnectionRequest: Protobuf ConnectionRequest.
        """
        request = ConnectionRequest()
        for address in self.addresses:
            address_info = request.addresses.add()
            address_info.host = address.host
            address_info.port = address.port
        request.tls_mode = TlsMode.SecureTls if self.use_tls else TlsMode.NoTls
        request.read_from = self.read_from.value
        if self.request_timeout:
            request.request_timeout = self.request_timeout
        if self.reconnect_strategy:
            request.connection_retry_strategy.number_of_retries = (
                self.reconnect_strategy.num_of_retries
            )
            request.connection_retry_strategy.factor = self.reconnect_strategy.factor
            request.connection_retry_strategy.exponent_base = (
                self.reconnect_strategy.exponent_base
            )
            if self.reconnect_strategy.jitter_percent is not None:
                request.connection_retry_strategy.jitter_percent = (
                    self.reconnect_strategy.jitter_percent
                )

        request.cluster_mode_enabled = True if cluster_mode else False
        if self.credentials:
            if self.credentials.username:
                request.authentication_info.username = self.credentials.username
            request.authentication_info.password = self.credentials.password
        if self.client_name:
            request.client_name = self.client_name
        request.protocol = self.protocol.value
        if self.inflight_requests_limit:
            request.inflight_requests_limit = self.inflight_requests_limit
        if self.client_az:
            request.client_az = self.client_az
        if self.advanced_config:
            self.advanced_config._create_a_protobuf_conn_request(request)

        return request

    def _is_pubsub_configured(self) -> bool:
        return False

    def _get_pubsub_callback_and_context(
        self,
    ) -> Tuple[Optional[Callable[[CoreCommands.PubSubMsg, Any], None]], Any]:
        return None, None


class AdvancedGlideClientConfiguration(AdvancedBaseClientConfiguration):
    """
    Represents the advanced configuration settings for a Standalone Glide client.
    """

    def __init__(self, connection_timeout: Optional[int] = None):

        super().__init__(connection_timeout)


class GlideClientConfiguration(BaseClientConfiguration):
    """
    Represents the configuration settings for a Standalone Glide client.

    Attributes:
        addresses (List[NodeAddress]): DNS Addresses and ports of known nodes in the cluster.
        Only nodes whose addresses were provided will be used by the client.
            For example::

                [
                    {address:sample-address-0001.use1.cache.amazonaws.com, port:6379},
                    {address: sample-address-0002.use2.cache.amazonaws.com, port:6379}
                ]

        use_tls (bool): True if communication with the cluster should use Transport Level Security.
        credentials (ServerCredentials): Credentials for authentication process.
                If none are set, the client will not authenticate itself with the server.
        read_from (ReadFrom): If not set, `PRIMARY` will be used.
        request_timeout (Optional[int]):  The duration in milliseconds that the client should wait for a request to complete.
                This duration encompasses sending the request, awaiting for a response from the server, and any required
                reconnections or retries.
                If the specified timeout is exceeded for a pending request, it will result in a timeout error.
                If not explicitly set, a default value of 250 milliseconds will be used.
        reconnect_strategy (Optional[BackoffStrategy]): Strategy used to determine how and when to reconnect, in case of
            connection failures.
            If not set, a default backoff strategy will be used.
        database_id (Optional[int]): index of the logical database to connect to.
        client_name (Optional[str]): Client name to be used for the client. Will be used with CLIENT SETNAME command during
            connection establishment.
        protocol (ProtocolVersion): The version of the RESP protocol to communicate with the server.
        pubsub_subscriptions (Optional[GlideClientConfiguration.PubSubSubscriptions]): Pubsub subscriptions to be used for the
                client.
                Will be applied via SUBSCRIBE/PSUBSCRIBE commands during connection establishment.
        inflight_requests_limit (Optional[int]): The maximum number of concurrent requests allowed to be in-flight
            (sent but not yet completed).
            This limit is used to control the memory usage and prevent the client from overwhelming the server or getting
            stuck in case of a queue backlog.
            If not set, a default value will be used.
        client_az (Optional[str]): Availability Zone of the client.
            If ReadFrom strategy is AZAffinity, this setting ensures that readonly commands are directed to replicas within
            the specified AZ if exits.
            If ReadFrom strategy is AZAffinityReplicasAndPrimary, this setting ensures that readonly commands are directed to
            nodes (first replicas then primary) within the specified AZ if they exist.
        advanced_config (Optional[AdvancedGlideClientConfiguration]): Advanced configuration settings for the client,
            see `AdvancedGlideClientConfiguration`.
    """

    class PubSubChannelModes(IntEnum):
        """
        Describes pubsub subsciption modes.
        See [valkey.io](https://valkey.io/docs/topics/pubsub/) for more details
        """

        Exact = 0
        """ Use exact channel names """
        Pattern = 1
        """ Use channel name patterns """

    @dataclass
    class PubSubSubscriptions:
        """Describes pubsub configuration for standalone mode client.

        Attributes:
            channels_and_patterns (Dict[GlideClientConfiguration.PubSubChannelModes, Set[str]]):
                Channels and patterns by modes.
            callback (Optional[Callable[[CoreCommands.PubSubMsg, Any], None]]):
                Optional callback to accept the incomming messages.
            context (Any):
                Arbitrary context to pass to the callback.
        """

        channels_and_patterns: Dict[
            GlideClientConfiguration.PubSubChannelModes, Set[str]
        ]
        callback: Optional[Callable[[CoreCommands.PubSubMsg, Any], None]]
        context: Any

    def __init__(
        self,
        addresses: List[NodeAddress],
        use_tls: bool = False,
        credentials: Optional[ServerCredentials] = None,
        read_from: ReadFrom = ReadFrom.PRIMARY,
        request_timeout: Optional[int] = None,
        reconnect_strategy: Optional[BackoffStrategy] = None,
        database_id: Optional[int] = None,
        client_name: Optional[str] = None,
        protocol: ProtocolVersion = ProtocolVersion.RESP3,
        pubsub_subscriptions: Optional[PubSubSubscriptions] = None,
        inflight_requests_limit: Optional[int] = None,
        client_az: Optional[str] = None,
        advanced_config: Optional[AdvancedGlideClientConfiguration] = None,
    ):
        super().__init__(
            addresses=addresses,
            use_tls=use_tls,
            credentials=credentials,
            read_from=read_from,
            request_timeout=request_timeout,
            reconnect_strategy=reconnect_strategy,
            client_name=client_name,
            protocol=protocol,
            inflight_requests_limit=inflight_requests_limit,
            client_az=client_az,
            advanced_config=advanced_config,
        )
        self.database_id = database_id
        self.pubsub_subscriptions = pubsub_subscriptions

    def _create_a_protobuf_conn_request(
        self, cluster_mode: bool = False
    ) -> ConnectionRequest:
        assert cluster_mode is False
        request = super()._create_a_protobuf_conn_request(cluster_mode)
        if self.database_id:
            request.database_id = self.database_id

        if self.pubsub_subscriptions:
            if self.protocol == ProtocolVersion.RESP2:
                raise ConfigurationError(
                    "PubSub subscriptions require RESP3 protocol, but RESP2 was configured."
                )
            if (
                self.pubsub_subscriptions.context is not None
                and not self.pubsub_subscriptions.callback
            ):
                raise ConfigurationError(
                    "PubSub subscriptions with a context require a callback function to be configured."
                )
            for (
                channel_type,
                channels_patterns,
            ) in self.pubsub_subscriptions.channels_and_patterns.items():
                entry = request.pubsub_subscriptions.channels_or_patterns_by_type[
                    int(channel_type)
                ]
                for channel_pattern in channels_patterns:
                    entry.channels_or_patterns.append(str.encode(channel_pattern))

        return request

    def _is_pubsub_configured(self) -> bool:
        return self.pubsub_subscriptions is not None

    def _get_pubsub_callback_and_context(
        self,
    ) -> Tuple[Optional[Callable[[CoreCommands.PubSubMsg, Any], None]], Any]:
        if self.pubsub_subscriptions:
            return self.pubsub_subscriptions.callback, self.pubsub_subscriptions.context
        return None, None


class AdvancedGlideClusterClientConfiguration(AdvancedBaseClientConfiguration):
    """
    Represents the advanced configuration settings for a Glide Cluster client.
    """

    def __init__(self, connection_timeout: Optional[int] = None):
        super().__init__(connection_timeout)


class GlideClusterClientConfiguration(BaseClientConfiguration):
    """
    Represents the configuration settings for a Cluster Glide client.

    Attributes:
        addresses (List[NodeAddress]): DNS Addresses and ports of known nodes in the cluster.
            The list can be partial, as the client will attempt to map out the cluster and find all nodes.
            For example::

                [
                    {address:configuration-endpoint.use1.cache.amazonaws.com, port:6379}
                ]

        use_tls (bool): True if communication with the cluster should use Transport Level Security.
        credentials (ServerCredentials): Credentials for authentication process.
                If none are set, the client will not authenticate itself with the server.
        read_from (ReadFrom): If not set, `PRIMARY` will be used.
        request_timeout (Optional[int]):  The duration in milliseconds that the client should wait for a request to complete.
            This duration encompasses sending the request, awaiting for a response from the server, and any required
            reconnections or retries.
            If the specified timeout is exceeded for a pending request, it will result in a timeout error. If not explicitly
            set, a default value of 250 milliseconds will be used.
        reconnect_strategy (Optional[BackoffStrategy]): Strategy used to determine how and when to reconnect, in case of
            connection failures.
            If not set, a default backoff strategy will be used.
        client_name (Optional[str]): Client name to be used for the client. Will be used with CLIENT SETNAME command during
            connection establishment.
        protocol (ProtocolVersion): The version of the RESP protocol to communicate with the server.
        periodic_checks (Union[PeriodicChecksStatus, PeriodicChecksManualInterval]): Configure the periodic topology checks.
            These checks evaluate changes in the cluster's topology, triggering a slot refresh when detected.
            Periodic checks ensure a quick and efficient process by querying a limited number of nodes.
            Defaults to PeriodicChecksStatus.ENABLED_DEFAULT_CONFIGS.
        pubsub_subscriptions (Optional[GlideClusterClientConfiguration.PubSubSubscriptions]): Pubsub subscriptions to be used
            for the client.
            Will be applied via SUBSCRIBE/PSUBSCRIBE/SSUBSCRIBE commands during connection establishment.
        inflight_requests_limit (Optional[int]): The maximum number of concurrent requests allowed to be in-flight
            (sent but not yet completed).
            This limit is used to control the memory usage and prevent the client from overwhelming the server or getting
            stuck in case of a queue backlog.
            If not set, a default value will be used.
        client_az (Optional[str]): Availability Zone of the client.
            If ReadFrom strategy is AZAffinity, this setting ensures that readonly commands are directed to replicas within
            the specified AZ if exits.
            If ReadFrom strategy is AZAffinityReplicasAndPrimary, this setting ensures that readonly commands are directed to
            nodes (first replicas then primary) within the specified AZ if they exist.
        advanced_config (Optional[AdvancedGlideClusterClientConfiguration]) : Advanced configuration settings for the client,
            see `AdvancedGlideClusterClientConfiguration`.


    Note:
        Currently, the reconnection strategy in cluster mode is not configurable, and exponential backoff
        with fixed values is used.
    """

    class PubSubChannelModes(IntEnum):
        """
        Describes pubsub subsciption modes.
        See [valkey.io](https://valkey.io/docs/topics/pubsub/) for more details
        """

        Exact = 0
        """ Use exact channel names """
        Pattern = 1
        """ Use channel name patterns """
        Sharded = 2
        """ Use sharded pubsub. Available since Valkey version 7.0. """

    @dataclass
    class PubSubSubscriptions:
        """Describes pubsub configuration for cluster mode client.

        Attributes:
            channels_and_patterns (Dict[GlideClusterClientConfiguration.PubSubChannelModes, Set[str]]):
                Channels and patterns by modes.
            callback (Optional[Callable[[CoreCommands.PubSubMsg, Any], None]]):
                Optional callback to accept the incoming messages.
            context (Any):
                Arbitrary context to pass to the callback.
        """

        channels_and_patterns: Dict[
            GlideClusterClientConfiguration.PubSubChannelModes, Set[str]
        ]
        callback: Optional[Callable[[CoreCommands.PubSubMsg, Any], None]]
        context: Any

    def __init__(
        self,
        addresses: List[NodeAddress],
        use_tls: bool = False,
        credentials: Optional[ServerCredentials] = None,
        read_from: ReadFrom = ReadFrom.PRIMARY,
        request_timeout: Optional[int] = None,
        reconnect_strategy: Optional[BackoffStrategy] = None,
        client_name: Optional[str] = None,
        protocol: ProtocolVersion = ProtocolVersion.RESP3,
        periodic_checks: Union[
            PeriodicChecksStatus, PeriodicChecksManualInterval
        ] = PeriodicChecksStatus.ENABLED_DEFAULT_CONFIGS,
        pubsub_subscriptions: Optional[PubSubSubscriptions] = None,
        inflight_requests_limit: Optional[int] = None,
        client_az: Optional[str] = None,
        advanced_config: Optional[AdvancedGlideClusterClientConfiguration] = None,
    ):
        super().__init__(
            addresses=addresses,
            use_tls=use_tls,
            credentials=credentials,
            read_from=read_from,
            request_timeout=request_timeout,
            reconnect_strategy=reconnect_strategy,
            client_name=client_name,
            protocol=protocol,
            inflight_requests_limit=inflight_requests_limit,
            client_az=client_az,
            advanced_config=advanced_config,
        )
        self.periodic_checks = periodic_checks
        self.pubsub_subscriptions = pubsub_subscriptions

    def _create_a_protobuf_conn_request(
        self, cluster_mode: bool = False
    ) -> ConnectionRequest:
        assert cluster_mode is True
        request = super()._create_a_protobuf_conn_request(cluster_mode)
        if type(self.periodic_checks) is PeriodicChecksManualInterval:
            request.periodic_checks_manual_interval.duration_in_sec = (
                self.periodic_checks.duration_in_sec
            )
        elif self.periodic_checks == PeriodicChecksStatus.DISABLED:
            request.periodic_checks_disabled.SetInParent()

        if self.pubsub_subscriptions:
            if self.protocol == ProtocolVersion.RESP2:
                raise ConfigurationError(
                    "PubSub subscriptions require RESP3 protocol, but RESP2 was configured."
                )
            if (
                self.pubsub_subscriptions.context is not None
                and not self.pubsub_subscriptions.callback
            ):
                raise ConfigurationError(
                    "PubSub subscriptions with a context require a callback function to be configured."
                )
            for (
                channel_type,
                channels_patterns,
            ) in self.pubsub_subscriptions.channels_and_patterns.items():
                entry = request.pubsub_subscriptions.channels_or_patterns_by_type[
                    int(channel_type)
                ]
                for channel_pattern in channels_patterns:
                    entry.channels_or_patterns.append(str.encode(channel_pattern))

        return request

    def _is_pubsub_configured(self) -> bool:
        return self.pubsub_subscriptions is not None

    def _get_pubsub_callback_and_context(
        self,
    ) -> Tuple[Optional[Callable[[CoreCommands.PubSubMsg, Any], None]], Any]:
        if self.pubsub_subscriptions:
            return self.pubsub_subscriptions.callback, self.pubsub_subscriptions.context
        return None, None



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from typing import List, Mapping, Optional

from glide.async_commands.server_modules.ft_options.ft_constants import FtSearchKeywords
from glide.constants import TEncodable


class FtSearchLimit:
    """
    This class represents the arguments for the LIMIT option of the FT.SEARCH command.
    """

    def __init__(self, offset: int, count: int):
        """
        Initialize a new FtSearchLimit instance.

        Args:
            offset (int): The number of keys to skip before returning the result for the FT.SEARCH command.
            count (int): The total number of keys to be returned by FT.SEARCH command.
        """
        self.offset = offset
        self.count = count

    def to_args(self) -> List[TEncodable]:
        """
        Get the arguments for the LIMIT option of FT.SEARCH.

        Returns:
            List[TEncodable]: A list of LIMIT option arguments.
        """
        args: List[TEncodable] = [
            FtSearchKeywords.LIMIT,
            str(self.offset),
            str(self.count),
        ]
        return args


class ReturnField:
    """
    This class represents the arguments for the RETURN option of the FT.SEARCH command.
    """

    def __init__(
        self, field_identifier: TEncodable, alias: Optional[TEncodable] = None
    ):
        """
        Initialize a new ReturnField instance.

        Args:
            field_identifier (TEncodable): The identifier for the field of the key that has to returned as a result of
                FT.SEARCH command.
            alias (Optional[TEncodable]): The alias to override the name of the field in the FT.SEARCH result.
        """
        self.field_identifier = field_identifier
        self.alias = alias

    def to_args(self) -> List[TEncodable]:
        """
        Get the arguments for the RETURN option of FT.SEARCH.

        Returns:
            List[TEncodable]: A list of RETURN option arguments.
        """
        args: List[TEncodable] = [self.field_identifier]
        if self.alias:
            args.append(FtSearchKeywords.AS)
            args.append(self.alias)
        return args


class FtSearchOptions:
    """
    This class represents the input options to be used in the FT.SEARCH command.
    All fields in this class are optional inputs for FT.SEARCH.
    """

    def __init__(
        self,
        return_fields: Optional[List[ReturnField]] = None,
        timeout: Optional[int] = None,
        params: Optional[Mapping[TEncodable, TEncodable]] = None,
        limit: Optional[FtSearchLimit] = None,
        count: Optional[bool] = False,
    ):
        """
        Initialize the FT.SEARCH optional fields.

        Args:
            return_fields (Optional[List[ReturnField]]): The fields of a key that are returned by FT.SEARCH command.
                See `ReturnField`.
            timeout (Optional[int]): This value overrides the timeout parameter of the module.
                The unit for the timout is in milliseconds.
            params (Optional[Mapping[TEncodable, TEncodable]]): Param key/value pairs that can be referenced from within the
                query expression.
            limit (Optional[FtSearchLimit]): This option provides pagination capability. Only the keys that satisfy the offset
                and count values are returned. See `FtSearchLimit`.
            count (Optional[bool]): This flag option suppresses returning the contents of keys.
                Only the number of keys is returned.
        """
        self.return_fields = return_fields
        self.timeout = timeout
        self.params = params
        self.limit = limit
        self.count = count

    def to_args(self) -> List[TEncodable]:
        """
        Get the optional arguments for the FT.SEARCH command.

        Returns:
            List[TEncodable]:
                List of FT.SEARCH optional agruments.
        """
        args: List[TEncodable] = []
        if self.return_fields:
            args.append(FtSearchKeywords.RETURN)
            return_field_args: List[TEncodable] = []
            for return_field in self.return_fields:
                return_field_args.extend(return_field.to_args())
            args.append(str(len(return_field_args)))
            args.extend(return_field_args)
        if self.timeout:
            args.append(FtSearchKeywords.TIMEOUT)
            args.append(str(self.timeout))
        if self.params:
            args.append(FtSearchKeywords.PARAMS)
            args.append(str(len(self.params) * 2))
            for name, value in self.params.items():
                args.append(name)
                args.append(value)
        if self.limit:
            args.extend(self.limit.to_args())
        if self.count:
            args.append(FtSearchKeywords.COUNT)
        return args



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import asyncio
import sys
import threading
from typing import Any, Dict, List, Optional, Tuple, Type, Union, cast

from glide.async_commands.cluster_commands import ClusterCommands
from glide.async_commands.command_args import ObjectType
from glide.async_commands.core import CoreCommands
from glide.async_commands.standalone_commands import StandaloneCommands
from glide.config import BaseClientConfiguration, ServerCredentials
from glide.constants import DEFAULT_READ_BYTES_SIZE, OK, TEncodable, TRequest, TResult
from glide.exceptions import (
    ClosingError,
    ConfigurationError,
    ConnectionError,
    ExecAbortError,
    RequestError,
    TimeoutError,
)
from glide.logger import Level as LogLevel
from glide.logger import Logger as ClientLogger
from glide.protobuf.command_request_pb2 import Command, CommandRequest, RequestType
from glide.protobuf.connection_request_pb2 import ConnectionRequest
from glide.protobuf.response_pb2 import RequestErrorType, Response
from glide.protobuf_codec import PartialMessageException, ProtobufCodec
from glide.routes import Route, set_protobuf_route

from .glide import (
    DEFAULT_TIMEOUT_IN_MILLISECONDS,
    MAX_REQUEST_ARGS_LEN,
    ClusterScanCursor,
    create_leaked_bytes_vec,
    get_statistics,
    start_socket_listener_external,
    value_from_pointer,
)

if sys.version_info >= (3, 11):
    import asyncio as async_timeout
    from typing import Self
else:
    import async_timeout
    from typing_extensions import Self


def get_request_error_class(
    error_type: Optional[RequestErrorType.ValueType],
) -> Type[RequestError]:
    if error_type == RequestErrorType.Disconnect:
        return ConnectionError
    if error_type == RequestErrorType.ExecAbort:
        return ExecAbortError
    if error_type == RequestErrorType.Timeout:
        return TimeoutError
    if error_type == RequestErrorType.Unspecified:
        return RequestError
    return RequestError


class BaseClient(CoreCommands):
    def __init__(self, config: BaseClientConfiguration):
        """
        To create a new client, use the `create` classmethod
        """
        self.config: BaseClientConfiguration = config
        self._available_futures: Dict[int, asyncio.Future] = {}
        self._available_callback_indexes: List[int] = list()
        self._buffered_requests: List[TRequest] = list()
        self._writer_lock = threading.Lock()
        self.socket_path: Optional[str] = None
        self._reader_task: Optional[asyncio.Task] = None
        self._is_closed: bool = False
        self._pubsub_futures: List[asyncio.Future] = []
        self._pubsub_lock = threading.Lock()
        self._pending_push_notifications: List[Response] = list()

    @classmethod
    async def create(cls, config: BaseClientConfiguration) -> Self:
        """Creates a Glide client.

        Args:
            config (ClientConfiguration): The configuration options for the client, including cluster addresses,
            authentication credentials, TLS settings, periodic checks, and Pub/Sub subscriptions.

        Returns:
            Self: A promise that resolves to a connected client instance.

        Examples:
            # Connecting to a Standalone Server
            >>> from glide import GlideClientConfiguration, NodeAddress, GlideClient, ServerCredentials, BackoffStrategy
            >>> config = GlideClientConfiguration(
            ...     [
            ...         NodeAddress('primary.example.com', 6379),
            ...         NodeAddress('replica1.example.com', 6379),
            ...     ],
            ...     use_tls = True,
            ...     database_id = 1,
            ...     credentials = ServerCredentials(username = 'user1', password = 'passwordA'),
            ...     reconnect_strategy = BackoffStrategy(num_of_retries = 5, factor = 1000, exponent_base = 2),
            ...     pubsub_subscriptions = GlideClientConfiguration.PubSubSubscriptions(
            ...         channels_and_patterns = {GlideClientConfiguration.PubSubChannelModes.Exact: {'updates'}},
            ...         callback = lambda message,context : print(message),
            ...     ),
            ... )
            >>> client = await GlideClient.create(config)

            # Connecting to a Cluster
            >>> from glide import GlideClusterClientConfiguration, NodeAddress, GlideClusterClient,
            ... PeriodicChecksManualInterval
            >>> config = GlideClusterClientConfiguration(
            ...     [
            ...         NodeAddress('address1.example.com', 6379),
            ...         NodeAddress('address2.example.com', 6379),
            ...     ],
            ...     use_tls = True,
            ...     periodic_checks = PeriodicChecksManualInterval(duration_in_sec = 30),
            ...     credentials = ServerCredentials(username = 'user1', password = 'passwordA'),
            ...     reconnect_strategy = BackoffStrategy(num_of_retries = 5, factor = 1000, exponent_base = 2),
            ...     pubsub_subscriptions = GlideClusterClientConfiguration.PubSubSubscriptions(
            ...         channels_and_patterns = {
            ...             GlideClusterClientConfiguration.PubSubChannelModes.Exact: {'updates'},
            ...             GlideClusterClientConfiguration.PubSubChannelModes.Sharded: {'sharded_channel'},
            ...         },
            ...         callback = lambda message,context : print(message),
            ...     ),
            ... )
            >>> client = await GlideClusterClient.create(config)

        Remarks:
            Use this static method to create and connect a client to a Valkey server.
            The client will automatically handle connection establishment, including cluster topology discovery and
            handling of authentication and TLS configurations.

                - **Cluster Topology Discovery**: The client will automatically discover the cluster topology based
                  on the seed addresses provided.
                - **Authentication**: If `ServerCredentials` are provided, the client will attempt to authenticate
                  using the specified username and password.
                - **TLS**: If `use_tls` is set to `true`, the client will establish secure connections using TLS.
                - **Periodic Checks**: The `periodic_checks` setting allows you to configure how often the client
                  checks for cluster topology changes.
                - **Reconnection Strategy**: The `BackoffStrategy` settings define how the client will attempt to
                  reconnect in case of disconnections.
                - **Pub/Sub Subscriptions**: Any channels or patterns specified in `PubSubSubscriptions` will be
                  subscribed to upon connection.

        """
        config = config
        self = cls(config)
        init_future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()

        def init_callback(socket_path: Optional[str], err: Optional[str]):
            if err is not None:
                raise ClosingError(err)
            elif socket_path is None:
                raise ClosingError(
                    "Socket initialization error: Missing valid socket path."
                )
            else:
                # Received socket path
                self.socket_path = socket_path
                loop.call_soon_threadsafe(init_future.set_result, True)

        start_socket_listener_external(init_callback=init_callback)

        # will log if the logger was created (wrapper or costumer) on info
        # level or higher
        ClientLogger.log(LogLevel.INFO, "connection info", "new connection established")
        # Wait for the socket listener to complete its initialization
        await init_future
        # Create UDS connection
        await self._create_uds_connection()
        # Start the reader loop as a background task
        self._reader_task = asyncio.create_task(self._reader_loop())
        # Set the client configurations
        await self._set_connection_configurations()
        return self

    async def _create_uds_connection(self) -> None:
        try:
            # Open an UDS connection
            async with async_timeout.timeout(DEFAULT_TIMEOUT_IN_MILLISECONDS):
                reader, writer = await asyncio.open_unix_connection(
                    path=self.socket_path
                )
            self._reader = reader
            self._writer = writer
        except Exception as e:
            await self.close(f"Failed to create UDS connection: {e}")
            raise

    def __del__(self) -> None:
        try:
            if self._reader_task:
                self._reader_task.cancel()
        except RuntimeError as e:
            if "no running event loop" in str(e):
                # event loop already closed
                pass

    async def close(self, err_message: Optional[str] = None) -> None:
        """
        Terminate the client by closing all associated resources, including the socket and any active futures.
        All open futures will be closed with an exception.

        Args:
            err_message (Optional[str]): If not None, this error message will be passed along with the exceptions when
            closing all open futures.
            Defaults to None.
        """
        self._is_closed = True
        for response_future in self._available_futures.values():
            if not response_future.done():
                err_message = "" if err_message is None else err_message
                response_future.set_exception(ClosingError(err_message))
        try:
            self._pubsub_lock.acquire()
            for pubsub_future in self._pubsub_futures:
                if not pubsub_future.done() and not pubsub_future.cancelled():
                    pubsub_future.set_exception(ClosingError(""))
        finally:
            self._pubsub_lock.release()

        self._writer.close()
        await self._writer.wait_closed()
        self.__del__()

    def _get_future(self, callback_idx: int) -> asyncio.Future:
        response_future: asyncio.Future = asyncio.Future()
        self._available_futures.update({callback_idx: response_future})
        return response_future

    def _get_protobuf_conn_request(self) -> ConnectionRequest:
        return self.config._create_a_protobuf_conn_request()

    async def _set_connection_configurations(self) -> None:
        conn_request = self._get_protobuf_conn_request()
        response_future: asyncio.Future = self._get_future(0)
        await self._write_or_buffer_request(conn_request)
        await response_future
        if response_future.result() is not OK:
            raise ClosingError(response_future.result())

    def _create_write_task(self, request: TRequest):
        asyncio.create_task(self._write_or_buffer_request(request))

    async def _write_or_buffer_request(self, request: TRequest):
        self._buffered_requests.append(request)
        if self._writer_lock.acquire(False):
            try:
                while len(self._buffered_requests) > 0:
                    await self._write_buffered_requests_to_socket()

            finally:
                self._writer_lock.release()

    async def _write_buffered_requests_to_socket(self) -> None:
        requests = self._buffered_requests
        self._buffered_requests = list()
        b_arr = bytearray()
        for request in requests:
            ProtobufCodec.encode_delimited(b_arr, request)
        self._writer.write(b_arr)
        await self._writer.drain()

    def _encode_arg(self, arg: TEncodable) -> bytes:
        """
        Converts a string argument to bytes.

        Args:
            arg (str): An encodable argument.

        Returns:
            bytes: The encoded argument as bytes.
        """
        if isinstance(arg, str):
            # TODO: Allow passing different encoding options
            return bytes(arg, encoding="utf8")
        return arg

    def _encode_and_sum_size(
        self,
        args_list: Optional[List[TEncodable]],
    ) -> Tuple[List[bytes], int]:
        """
        Encodes the list and calculates the total memory size.

        Args:
            args_list (Optional[List[TEncodable]]): A list of strings to be converted to bytes.
                                                           If None or empty, returns ([], 0).

        Returns:
            int: The total memory size of the encoded arguments in bytes.
        """
        args_size = 0
        encoded_args_list: List[bytes] = []
        if not args_list:
            return (encoded_args_list, args_size)
        for arg in args_list:
            encoded_arg = self._encode_arg(arg) if isinstance(arg, str) else arg
            encoded_args_list.append(encoded_arg)
            args_size += len(encoded_arg)
        return (encoded_args_list, args_size)

    async def _execute_command(
        self,
        request_type: RequestType.ValueType,
        args: List[TEncodable],
        route: Optional[Route] = None,
    ) -> TResult:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )
        request = CommandRequest()
        request.callback_idx = self._get_callback_index()
        request.single_command.request_type = request_type
        request.single_command.args_array.args[:] = [
            bytes(elem, encoding="utf8") if isinstance(elem, str) else elem
            for elem in args
        ]
        (encoded_args, args_size) = self._encode_and_sum_size(args)
        if args_size < MAX_REQUEST_ARGS_LEN:
            request.single_command.args_array.args[:] = encoded_args
        else:
            request.single_command.args_vec_pointer = create_leaked_bytes_vec(
                encoded_args
            )
        set_protobuf_route(request, route)
        return await self._write_request_await_response(request)

    async def _execute_batch(
        self,
        commands: List[Tuple[RequestType.ValueType, List[TEncodable]]],
        is_atomic: bool,
        raise_on_error: bool = False,
        retry_server_error: bool = False,
        retry_connection_error: bool = False,
        route: Optional[Route] = None,
        timeout: Optional[int] = None,
    ) -> List[TResult]:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )
        request = CommandRequest()
        request.callback_idx = self._get_callback_index()
        batch_commands = []
        for requst_type, args in commands:
            command = Command()
            command.request_type = requst_type
            # For now, we allow the user to pass the command as array of strings
            # we convert them here into bytes (the datatype that our rust core expects)
            (encoded_args, args_size) = self._encode_and_sum_size(args)
            if args_size < MAX_REQUEST_ARGS_LEN:
                command.args_array.args[:] = encoded_args
            else:
                command.args_vec_pointer = create_leaked_bytes_vec(encoded_args)
            batch_commands.append(command)
        request.batch.commands.extend(batch_commands)
        request.batch.is_atomic = is_atomic
        request.batch.raise_on_error = raise_on_error
        if timeout is not None:
            request.batch.timeout = timeout
        request.batch.retry_server_error = retry_server_error
        request.batch.retry_connection_error = retry_connection_error
        set_protobuf_route(request, route)
        return await self._write_request_await_response(request)

    async def _execute_script(
        self,
        hash: str,
        keys: Optional[List[Union[str, bytes]]] = None,
        args: Optional[List[Union[str, bytes]]] = None,
        route: Optional[Route] = None,
    ) -> TResult:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )
        request = CommandRequest()
        request.callback_idx = self._get_callback_index()
        (encoded_keys, keys_size) = self._encode_and_sum_size(keys)
        (encoded_args, args_size) = self._encode_and_sum_size(args)
        if (keys_size + args_size) < MAX_REQUEST_ARGS_LEN:
            request.script_invocation.hash = hash
            request.script_invocation.keys[:] = encoded_keys
            request.script_invocation.args[:] = encoded_args

        else:
            request.script_invocation_pointers.hash = hash
            request.script_invocation_pointers.keys_pointer = create_leaked_bytes_vec(
                encoded_keys
            )
            request.script_invocation_pointers.args_pointer = create_leaked_bytes_vec(
                encoded_args
            )
        set_protobuf_route(request, route)
        return await self._write_request_await_response(request)

    async def get_pubsub_message(self) -> CoreCommands.PubSubMsg:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )

        if not self.config._is_pubsub_configured():
            raise ConfigurationError(
                "The operation will never complete since there was no pubsub subscriptions applied to the client."
            )

        if self.config._get_pubsub_callback_and_context()[0] is not None:
            raise ConfigurationError(
                "The operation will never complete since messages will be passed to the configured callback."
            )

        # locking might not be required
        response_future: asyncio.Future = asyncio.Future()
        try:
            self._pubsub_lock.acquire()
            self._pubsub_futures.append(response_future)
            self._complete_pubsub_futures_safe()
        finally:
            self._pubsub_lock.release()
        return await response_future

    def try_get_pubsub_message(self) -> Optional[CoreCommands.PubSubMsg]:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )

        if not self.config._is_pubsub_configured():
            raise ConfigurationError(
                "The operation will never succeed since there was no pubsbub subscriptions applied to the client."
            )

        if self.config._get_pubsub_callback_and_context()[0] is not None:
            raise ConfigurationError(
                "The operation will never succeed since messages will be passed to the configured callback."
            )

        # locking might not be required
        msg: Optional[CoreCommands.PubSubMsg] = None
        try:
            self._pubsub_lock.acquire()
            self._complete_pubsub_futures_safe()
            while len(self._pending_push_notifications) and not msg:
                push_notification = self._pending_push_notifications.pop(0)
                msg = self._notification_to_pubsub_message_safe(push_notification)
        finally:
            self._pubsub_lock.release()
        return msg

    def _cancel_pubsub_futures_with_exception_safe(self, exception: ConnectionError):
        while len(self._pubsub_futures):
            next_future = self._pubsub_futures.pop(0)
            if not next_future.cancelled():
                next_future.set_exception(exception)

    def _notification_to_pubsub_message_safe(
        self, response: Response
    ) -> Optional[CoreCommands.PubSubMsg]:
        pubsub_message = None
        push_notification = cast(
            Dict[str, Any], value_from_pointer(response.resp_pointer)
        )
        message_kind = push_notification["kind"]
        if message_kind == "Disconnection":
            ClientLogger.log(
                LogLevel.WARN,
                "disconnect notification",
                "Transport disconnected, messages might be lost",
            )
        elif (
            message_kind == "Message"
            or message_kind == "PMessage"
            or message_kind == "SMessage"
        ):
            values: List = push_notification["values"]
            if message_kind == "PMessage":
                pubsub_message = BaseClient.PubSubMsg(
                    message=values[2], channel=values[1], pattern=values[0]
                )
            else:
                pubsub_message = BaseClient.PubSubMsg(
                    message=values[1], channel=values[0], pattern=None
                )
        elif (
            message_kind == "PSubscribe"
            or message_kind == "Subscribe"
            or message_kind == "SSubscribe"
            or message_kind == "Unsubscribe"
            or message_kind == "PUnsubscribe"
            or message_kind == "SUnsubscribe"
        ):
            pass
        else:
            ClientLogger.log(
                LogLevel.WARN,
                "unknown notification",
                f"Unknown notification message: '{message_kind}'",
            )

        return pubsub_message

    def _complete_pubsub_futures_safe(self):
        while len(self._pending_push_notifications) and len(self._pubsub_futures):
            next_push_notification = self._pending_push_notifications.pop(0)
            pubsub_message = self._notification_to_pubsub_message_safe(
                next_push_notification
            )
            if pubsub_message:
                self._pubsub_futures.pop(0).set_result(pubsub_message)

    async def _write_request_await_response(self, request: CommandRequest):
        # Create a response future for this request and add it to the available
        # futures map
        response_future = self._get_future(request.callback_idx)
        self._create_write_task(request)
        await response_future
        return response_future.result()

    def _get_callback_index(self) -> int:
        try:
            return self._available_callback_indexes.pop()
        except IndexError:
            # The list is empty
            return len(self._available_futures)

    async def _process_response(self, response: Response) -> None:
        res_future = self._available_futures.pop(response.callback_idx, None)
        if not res_future or response.HasField("closing_error"):
            err_msg = (
                response.closing_error
                if response.HasField("closing_error")
                else f"Client Error - closing due to unknown error. callback index:  {response.callback_idx}"
            )
            if res_future is not None:
                res_future.set_exception(ClosingError(err_msg))
            await self.close(err_msg)
            raise ClosingError(err_msg)
        else:
            self._available_callback_indexes.append(response.callback_idx)
            if response.HasField("request_error"):
                error_type = get_request_error_class(response.request_error.type)
                res_future.set_exception(error_type(response.request_error.message))
            elif response.HasField("resp_pointer"):
                res_future.set_result(value_from_pointer(response.resp_pointer))
            elif response.HasField("constant_response"):
                res_future.set_result(OK)
            else:
                res_future.set_result(None)

    async def _process_push(self, response: Response) -> None:
        if response.HasField("closing_error") or not response.HasField("resp_pointer"):
            err_msg = (
                response.closing_error
                if response.HasField("closing_error")
                else "Client Error - push notification without resp_pointer"
            )
            await self.close(err_msg)
            raise ClosingError(err_msg)

        try:
            self._pubsub_lock.acquire()
            callback, context = self.config._get_pubsub_callback_and_context()
            if callback:
                pubsub_message = self._notification_to_pubsub_message_safe(response)
                if pubsub_message:
                    callback(pubsub_message, context)
            else:
                self._pending_push_notifications.append(response)
                self._complete_pubsub_futures_safe()
        finally:
            self._pubsub_lock.release()

    async def _reader_loop(self) -> None:
        # Socket reader loop
        remaining_read_bytes = bytearray()
        while True:
            read_bytes = await self._reader.read(DEFAULT_READ_BYTES_SIZE)
            if len(read_bytes) == 0:
                err_msg = "The communication layer was unexpectedly closed."
                await self.close(err_msg)
                raise ClosingError(err_msg)
            read_bytes = remaining_read_bytes + bytearray(read_bytes)
            read_bytes_view = memoryview(read_bytes)
            offset = 0
            while offset <= len(read_bytes):
                try:
                    response, offset = ProtobufCodec.decode_delimited(
                        read_bytes, read_bytes_view, offset, Response
                    )
                except PartialMessageException:
                    # Received only partial response, break the inner loop
                    remaining_read_bytes = read_bytes[offset:]
                    break
                response = cast(Response, response)
                if response.is_push:
                    await self._process_push(response=response)
                else:
                    await self._process_response(response=response)

    async def get_statistics(self) -> dict:
        return get_statistics()

    async def _update_connection_password(
        self, password: Optional[str], immediate_auth: bool
    ) -> TResult:
        request = CommandRequest()
        request.callback_idx = self._get_callback_index()
        if password is not None:
            request.update_connection_password.password = password
        request.update_connection_password.immediate_auth = immediate_auth
        response = await self._write_request_await_response(request)
        # Update the client binding side password if managed to change core configuration password
        if response is OK:
            if self.config.credentials is None:
                self.config.credentials = ServerCredentials(password=password or "")
                self.config.credentials.password = password or ""
        return response


class GlideClusterClient(BaseClient, ClusterCommands):
    """
    Client used for connection to cluster servers.
    Use :func:`~BaseClient.create` to request a client.
    For full documentation, see
    [Valkey GLIDE Wiki](https://github.com/valkey-io/valkey-glide/wiki/Python-wrapper#cluster)
    """

    async def _cluster_scan(
        self,
        cursor: ClusterScanCursor,
        match: Optional[TEncodable] = None,
        count: Optional[int] = None,
        type: Optional[ObjectType] = None,
        allow_non_covered_slots: bool = False,
    ) -> List[Union[ClusterScanCursor, List[bytes]]]:
        if self._is_closed:
            raise ClosingError(
                "Unable to execute requests; the client is closed. Please create a new client."
            )
        request = CommandRequest()
        request.callback_idx = self._get_callback_index()
        # Take out the id string from the wrapping object
        cursor_string = cursor.get_cursor()
        request.cluster_scan.cursor = cursor_string
        request.cluster_scan.allow_non_covered_slots = allow_non_covered_slots
        if match is not None:
            request.cluster_scan.match_pattern = (
                self._encode_arg(match) if isinstance(match, str) else match
            )
        if count is not None:
            request.cluster_scan.count = count
        if type is not None:
            request.cluster_scan.object_type = type.value
        response = await self._write_request_await_response(request)
        return [ClusterScanCursor(bytes(response[0]).decode()), response[1]]

    def _get_protobuf_conn_request(self) -> ConnectionRequest:
        return self.config._create_a_protobuf_conn_request(cluster_mode=True)


class GlideClient(BaseClient, StandaloneCommands):
    """
    Client used for connection to standalone servers.
    Use :func:`~BaseClient.create` to request a client.
    For full documentation, see
    [Valkey GLIDE Wiki](https://github.com/valkey-io/valkey-glide/wiki/Python-wrapper#standalone)
    """


TGlideClient = Union[GlideClient, GlideClusterClient]



import asyncio
from typing import AsyncGenerator, List, cast

import pytest
from glide import ByAddressRoute
from glide.async_commands.command_args import ObjectType
from glide.config import ProtocolVersion
from glide.exceptions import RequestError
from glide.glide import ClusterScanCursor
from glide.glide_client import GlideClient, GlideClusterClient

from tests.conftest import create_client
from tests.utils.cluster import ValkeyCluster
from tests.utils.utils import get_random_string


# Helper function to get a number of nodes, and ask the cluster till we get the number of nodes
async def is_cluster_ready(client: GlideClusterClient, count: int) -> bool:
    # we allow max 20 seconds to get the nodes
    timeout = 20
    start_time = asyncio.get_event_loop().time()

    while True:
        if asyncio.get_event_loop().time() - start_time > timeout:
            return False

        cluster_info = await client.custom_command(["CLUSTER", "INFO"])
        cluster_info_map = {}

        if cluster_info:
            info_str = (
                cluster_info
                if isinstance(cluster_info, str)
                else (
                    cluster_info.decode()
                    if isinstance(cluster_info, bytes)
                    else str(cluster_info)
                )
            )
            cluster_info_lines = info_str.split("\n")
            cluster_info_lines = [line for line in cluster_info_lines if line]

            for line in cluster_info_lines:
                key, value = line.split(":")
                cluster_info_map[key.strip()] = value.strip()

            if (
                cluster_info_map.get("cluster_state") == "ok"
                and int(cluster_info_map.get("cluster_known_nodes", "0")) == count
            ):
                break

        await asyncio.sleep(2)

    return True


# The slots not covered testing is messing with the cluster by removing a node, and then scanning the cluster
# When a node is forgot its getting into a banned state for one minutes, so in order to bring back the cluster to normal state
# we need to wait for the node to be unbanned, and then we can continue with the tests
# In order to avoid the time wasting and the chance that the restoration will not happen, we will run the test on separate
# cluster
@pytest.fixture(scope="function")
async def function_scoped_cluster():
    """
    Function-scoped fixture to create a new cluster for each test invocation.
    """
    cluster = ValkeyCluster(
        tls=False, cluster_mode=True, shard_count=3, replica_count=0
    )
    yield cluster
    del cluster


# Since the cluster for slots covered is created separately, we need to create a client for the specific cluster
# The client is created with 100 timeout so looping over the keys with scan will return the error before we finish the loop
# otherwise the test will be flaky
@pytest.fixture(scope="function")
async def glide_client_scoped(
    request, function_scoped_cluster: ValkeyCluster, protocol: ProtocolVersion
) -> AsyncGenerator[GlideClusterClient, None]:
    """
    Get client for tests, adjusted to use the function-scoped cluster.
    """
    client = await create_client(
        request,
        True,
        valkey_cluster=function_scoped_cluster,
        protocol=protocol,
    )
    assert isinstance(client, GlideClusterClient)
    yield client
    await client.close()


@pytest.mark.asyncio
class TestScan:
    # Cluster scan tests
    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_simple(self, glide_client: GlideClusterClient):
        key = get_random_string(10)
        expected_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        expected_keys_encoded = map(lambda k: k.encode(), expected_keys)
        cursor = ClusterScanCursor()
        keys: List[str] = []
        while not cursor.is_finished():
            result = await glide_client.scan(cursor)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = cast(List[str], result[1])
            keys.extend(result_keys)

        assert set(expected_keys_encoded) == set(keys)

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_with_object_type_and_pattern(
        self, glide_client: GlideClusterClient
    ):
        key = get_random_string(10)
        expected_keys = [f"key-{key}-{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        unexpected_type_keys = [f"{key}-{i}" for i in range(100, 200)]
        for key in unexpected_type_keys:
            await glide_client.sadd(key, ["value"])
        encoded_unexpected_type_keys = map(lambda k: k.encode(), unexpected_type_keys)
        unexpected_pattern_keys = [f"{i}" for i in range(200, 300)]
        await glide_client.mset({k: "value" for k in unexpected_pattern_keys})
        encoded_unexpected_pattern_keys = map(
            lambda k: k.encode(), unexpected_pattern_keys
        )
        keys: List[str] = []
        cursor = ClusterScanCursor()
        while not cursor.is_finished():
            result = await glide_client.scan(
                cursor, match=b"key-*", type=ObjectType.STRING
            )
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = cast(List[str], result[1])
            keys.extend(result_keys)

        assert set(encoded_expected_keys) == set(keys)
        assert not set(encoded_unexpected_type_keys).intersection(set(keys))
        assert not set(encoded_unexpected_pattern_keys).intersection(set(keys))

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_with_count(self, glide_client: GlideClusterClient):
        key = get_random_string(10)
        expected_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        cursor = ClusterScanCursor()
        keys: List[str] = []
        successful_compared_scans = 0
        while not cursor.is_finished():
            result_of_1 = await glide_client.scan(cursor, count=1)
            cursor = cast(ClusterScanCursor, result_of_1[0])
            result_keys_of_1 = cast(List[str], result_of_1[1])
            keys.extend(result_keys_of_1)
            if cursor.is_finished():
                break
            result_of_100 = await glide_client.scan(cursor, count=100)
            cursor = cast(ClusterScanCursor, result_of_100[0])
            result_keys_of_100 = cast(List[str], result_of_100[1])
            keys.extend(result_keys_of_100)
            if len(result_keys_of_100) > len(result_keys_of_1):
                successful_compared_scans += 1

        assert set(encoded_expected_keys) == set(keys)
        assert successful_compared_scans > 0

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_with_match(self, glide_client: GlideClusterClient):
        unexpected_keys = [f"{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in unexpected_keys})
        encoded_unexpected_keys = map(lambda k: k.encode(), unexpected_keys)
        key = get_random_string(10)
        expected_keys = [f"key-{key}-{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        cursor = ClusterScanCursor()
        keys: List[str] = []
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, match="key-*")
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = cast(List[str], result[1])
            keys.extend(result_keys)
        assert set(encoded_expected_keys) == set(keys)
        assert not set(encoded_unexpected_keys).intersection(set(keys))

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    # We test whether the cursor is cleaned up after it is deleted, which we expect to happen when th GC is called
    async def test_cluster_scan_cleaning_cursor(self, glide_client: GlideClusterClient):
        key = get_random_string(10)
        await glide_client.mset({k: "value" for k in [f"{key}{i}" for i in range(100)]})
        cursor = cast(
            ClusterScanCursor, (await glide_client.scan(ClusterScanCursor()))[0]
        )
        cursor_string = cursor.get_cursor()
        del cursor
        new_cursor_with_same_id = ClusterScanCursor(cursor_string)
        with pytest.raises(RequestError) as e_info:
            await glide_client.scan(new_cursor_with_same_id)
        assert "Invalid scan_state_cursor id" in str(e_info.value)

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_all_types(self, glide_client: GlideClusterClient):
        # We test that the scan command work for all types of keys
        key = get_random_string(10)
        string_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in string_keys})
        encoded_string_keys = list(map(lambda k: k.encode(), string_keys))

        set_key = get_random_string(10)
        set_keys = [f"{set_key}{i}" for i in range(100, 200)]
        for key in set_keys:
            await glide_client.sadd(key, ["value"])
        encoded_set_keys = list(map(lambda k: k.encode(), set_keys))

        hash_key = get_random_string(10)
        hash_keys = [f"{hash_key}{i}" for i in range(200, 300)]
        for key in hash_keys:
            await glide_client.hset(key, {"field": "value"})
        encoded_hash_keys = list(map(lambda k: k.encode(), hash_keys))

        list_key = get_random_string(10)
        list_keys = [f"{list_key}{i}" for i in range(300, 400)]
        for key in list_keys:
            await glide_client.lpush(key, ["value"])
        encoded_list_keys = list(map(lambda k: k.encode(), list_keys))

        zset_key = get_random_string(10)
        zset_keys = [f"{zset_key}{i}" for i in range(400, 500)]
        for key in zset_keys:
            await glide_client.zadd(key, {"value": 1})
        encoded_zset_keys = list(map(lambda k: k.encode(), zset_keys))

        stream_key = get_random_string(10)
        stream_keys = [f"{stream_key}{i}" for i in range(500, 600)]
        for key in stream_keys:
            await glide_client.xadd(key, [("field", "value")])
        encoded_stream_keys = list(map(lambda k: k.encode(), stream_keys))

        cursor = ClusterScanCursor()
        keys: List[bytes] = []
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.STRING)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_string_keys) == set(keys)
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))

        cursor = ClusterScanCursor()
        keys.clear()
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.SET)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_set_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))

        cursor = ClusterScanCursor()
        keys.clear()
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.HASH)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_hash_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))

        cursor = ClusterScanCursor()
        keys.clear()
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.LIST)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_list_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))

        cursor = ClusterScanCursor()
        keys.clear()
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.ZSET)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_zset_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))

        cursor = ClusterScanCursor()
        keys.clear()
        while not cursor.is_finished():
            result = await glide_client.scan(cursor, type=ObjectType.STREAM)
            cursor = cast(ClusterScanCursor, result[0])
            result_keys = result[1]
            keys.extend(cast(List[bytes], result_keys))
        assert set(encoded_stream_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))

    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_cluster_scan_non_covered_slots(
        self,
        protocol: ProtocolVersion,
        function_scoped_cluster: ValkeyCluster,
        glide_client_scoped: GlideClusterClient,
    ):
        key = get_random_string(10)
        for i in range(1000):
            await glide_client_scoped.set(f"{key}{i}", "value")
        cursor = ClusterScanCursor()
        result = await glide_client_scoped.scan(cursor)
        cursor = cast(ClusterScanCursor, result[0])
        assert not cursor.is_finished()
        await glide_client_scoped.config_set({"cluster-require-full-coverage": "no"})
        # forget one server
        address_to_forget = glide_client_scoped.config.addresses[0]
        all_other_addresses = glide_client_scoped.config.addresses[1:]
        id_to_forget = await glide_client_scoped.custom_command(
            ["CLUSTER", "MYID"],
            ByAddressRoute(address_to_forget.host, address_to_forget.port),
        )
        for address in all_other_addresses:
            await glide_client_scoped.custom_command(
                ["CLUSTER", "FORGET", cast(str, id_to_forget)],
                ByAddressRoute(address.host, address.port),
            )
        # now we let it few seconds gossip to get the new cluster configuration
        await is_cluster_ready(glide_client_scoped, len(all_other_addresses))
        # Iterate scan until error is returned, as it might take time for the inner core to forget the missing node
        cursor = ClusterScanCursor()
        while True:
            try:
                while not cursor.is_finished():
                    result = await glide_client_scoped.scan(cursor)
                    cursor = cast(ClusterScanCursor, result[0])
                # Reset cursor for next iteration
                cursor = ClusterScanCursor()
            except RequestError as e_info:
                assert (
                    "Could not find an address covering a slot, SCAN operation cannot continue"
                    in str(e_info)
                )
                break
        # Scan with allow_non_covered_slots=True
        while not cursor.is_finished():
            result = await glide_client_scoped.scan(
                cursor, allow_non_covered_slots=True
            )
            cursor = cast(ClusterScanCursor, result[0])
        assert cursor.is_finished()
        # check the keys we can get with keys command, and scan from the beginning
        keys = []
        for address in all_other_addresses:
            result_keys = cast(
                List[bytes],
                await glide_client_scoped.custom_command(
                    ["KEYS", "*"], ByAddressRoute(address.host, address.port)
                ),
            )
            keys.extend(cast(List[bytes], result_keys))

        cursor = ClusterScanCursor()
        results = []
        while not cursor.is_finished():
            result = await glide_client_scoped.scan(
                cursor, allow_non_covered_slots=True
            )
            results.extend(cast(List[bytes], result[1]))
            cursor = cast(ClusterScanCursor, result[0])
        assert set(keys) == set(results)

    # Standalone scan tests
    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_standalone_scan_simple(self, glide_client: GlideClient):
        key = get_random_string(10)
        expected_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        keys: List[str] = []
        cursor = b"0"
        while True:
            result = await glide_client.scan(cursor)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes
            new_keys = cast(List[str], result[1])
            keys.extend(new_keys)
            if cursor == b"0":
                break
        assert set(encoded_expected_keys) == set(keys)

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_standalone_scan_with_object_type_and_pattern(
        self, glide_client: GlideClient
    ):
        key = get_random_string(10)
        expected_keys = [f"key-{key}-{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        unexpected_type_keys = [f"key-{i}" for i in range(100, 200)]
        for key in unexpected_type_keys:
            await glide_client.sadd(key, ["value"])
        unexpected_pattern_keys = [f"{i}" for i in range(200, 300)]
        for key in unexpected_pattern_keys:
            await glide_client.set(key, "value")
        keys: List[str] = []
        cursor = b"0"
        while True:
            result = await glide_client.scan(
                cursor, match=b"key-*", type=ObjectType.STRING
            )
            cursor = cast(bytes, result[0])
            keys.extend(list(map(lambda k: k.decode(), cast(List[bytes], result[1]))))
            if cursor == b"0":
                break
        assert set(expected_keys) == set(keys)
        assert not set(unexpected_type_keys).intersection(set(keys))
        assert not set(unexpected_pattern_keys).intersection(set(keys))

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_standalone_scan_with_count(self, glide_client: GlideClient):
        key = get_random_string(10)
        expected_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        cursor = "0"
        keys: List[str] = []
        successful_compared_scans = 0
        while True:
            result_of_1 = await glide_client.scan(cursor, count=1)
            cursor_bytes = cast(bytes, result_of_1[0])
            cursor = cursor_bytes.decode()
            keys_of_1 = cast(List[str], result_of_1[1])
            keys.extend(keys_of_1)
            result_of_100 = await glide_client.scan(cursor, count=100)
            cursor_bytes = cast(bytes, result_of_100[0])
            cursor = cursor_bytes.decode()
            keys_of_100 = cast(List[str], result_of_100[1])
            keys.extend(keys_of_100)
            if len(keys_of_100) > len(keys_of_1):
                successful_compared_scans += 1
            if cursor == "0":
                break
        assert set(encoded_expected_keys) == set(keys)
        assert successful_compared_scans > 0

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_standalone_scan_with_match(self, glide_client: GlideClient):
        key = get_random_string(10)
        expected_keys = [f"key-{key}-{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in expected_keys})
        encoded_expected_keys = map(lambda k: k.encode(), expected_keys)
        unexpected_keys = [f"{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in [f"{i}" for i in range(100)]})
        encoded_unexpected_keys = map(lambda k: k.encode(), unexpected_keys)
        cursor = "0"
        keys: List[str] = []
        while True:
            result = await glide_client.scan(cursor, match="key-*")
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = cast(List[str], result[1])
            keys.extend(new_keys)
            if cursor == "0":
                break
        assert set(encoded_expected_keys) == set(keys)
        assert not set(encoded_unexpected_keys).intersection(set(keys))

    @pytest.mark.parametrize("cluster_mode", [False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_standalone_scan_all_types(self, glide_client: GlideClient):
        # We test that the scan command work for all types of keys
        (
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
        ) = await self.create_keys(glide_client)

        cursor = "0"
        keys: List[bytes] = []
        cursor = await self.scan_string(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

        keys.clear()
        cursor = await self.scan_set(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

        keys.clear()
        cursor = await self.scan_hash(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

        keys.clear()
        cursor = await self.scan_list(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

        keys.clear()
        cursor = await self.scan_zset(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

        keys.clear()
        await self.scan_stream(
            glide_client,
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
            cursor,
            keys,
        )

    async def scan_stream(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.STREAM)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_stream_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))

    async def scan_zset(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.ZSET)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_zset_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))
        return cursor

    async def scan_list(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.LIST)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_list_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))
        return cursor

    async def scan_hash(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.HASH)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_hash_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))
        return cursor

    async def scan_set(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.SET)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_set_keys) == set(keys)
        assert not set(encoded_string_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))
        return cursor

    async def scan_string(
        self,
        glide_client,
        encoded_string_keys,
        encoded_set_keys,
        encoded_hash_keys,
        encoded_list_keys,
        encoded_zset_keys,
        encoded_stream_keys,
        cursor,
        keys,
    ):
        while True:
            result = await glide_client.scan(cursor, type=ObjectType.STRING)
            cursor_bytes = cast(bytes, result[0])
            cursor = cursor_bytes.decode()
            new_keys = result[1]
            keys.extend(cast(List[bytes], new_keys))
            if cursor == "0":
                break
        assert set(encoded_string_keys) == set(keys)
        assert not set(encoded_set_keys).intersection(set(keys))
        assert not set(encoded_hash_keys).intersection(set(keys))
        assert not set(encoded_list_keys).intersection(set(keys))
        assert not set(encoded_zset_keys).intersection(set(keys))
        assert not set(encoded_stream_keys).intersection(set(keys))
        return cursor

    async def create_keys(self, glide_client):
        key = get_random_string(10)
        string_keys = [f"{key}{i}" for i in range(100)]
        await glide_client.mset({k: "value" for k in string_keys})
        encoded_string_keys = list(map(lambda k: k.encode(), string_keys))

        set_keys = [f"{key}{i}" for i in range(100, 200)]
        for key in set_keys:
            await glide_client.sadd(key, ["value"])
        encoded_set_keys = list(map(lambda k: k.encode(), set_keys))

        hash_keys = [f"{key}{i}" for i in range(200, 300)]
        for key in hash_keys:
            await glide_client.hset(key, {"field": "value"})
        encoded_hash_keys = list(map(lambda k: k.encode(), hash_keys))

        list_keys = [f"{key}{i}" for i in range(300, 400)]
        for key in list_keys:
            await glide_client.lpush(key, ["value"])
        encoded_list_keys = list(map(lambda k: k.encode(), list_keys))

        zset_keys = [f"{key}{i}" for i in range(400, 500)]
        for key in zset_keys:
            await glide_client.zadd(key, {"value": 1})
        encoded_zset_keys = list(map(lambda k: k.encode(), zset_keys))

        stream_keys = [f"{key}{i}" for i in range(500, 600)]
        for key in stream_keys:
            await glide_client.xadd(key, [("field", "value")])
        encoded_stream_keys = list(map(lambda k: k.encode(), stream_keys))

        return (
            encoded_string_keys,
            encoded_set_keys,
            encoded_hash_keys,
            encoded_list_keys,
            encoded_zset_keys,
            encoded_stream_keys,
        )



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional

from glide.async_commands.server_modules.ft_options.ft_constants import FtCreateKeywords
from glide.constants import TEncodable


class FieldType(Enum):
    """
    All possible values for the data type of field identifier for the SCHEMA option.
    """

    TEXT = "TEXT"
    """
    If the field contains any blob of data.
    """
    TAG = "TAG"
    """
    If the field contains a tag field.
    """
    NUMERIC = "NUMERIC"
    """
    If the field contains a number.
    """
    VECTOR = "VECTOR"
    """
    If the field is a vector field that supports vector search.
    """


class VectorAlgorithm(Enum):
    """
    Algorithm for vector type fields used for vector similarity search.
    """

    HNSW = "HNSW"
    """
    Hierarchical Navigable Small World algorithm.
    """
    FLAT = "FLAT"
    """
    Flat algorithm or the brute force algorithm.
    """


class DistanceMetricType(Enum):
    """
    Distance metrics to measure the degree of similarity between two vectors.

    The above metrics calculate distance between two vectors, where the smaller the value is, the
    closer the two vectors are in the vector space.
    """

    L2 = "L2"
    """
    Euclidean distance
    """
    IP = "IP"
    """
    Inner product
    """
    COSINE = "COSINE"
    """
    Cosine distance
    """


class VectorType(Enum):
    """
    Type type for the vector field type.
    """

    FLOAT32 = "FLOAT32"
    """
    FLOAT32 type of vector. The only supported type.
    """


class Field(ABC):
    """
    Abstract base class for a vector search field.
    """

    @abstractmethod
    def __init__(
        self,
        name: TEncodable,
        type: FieldType,
        alias: Optional[TEncodable] = None,
    ):
        """
        Initialize a new field instance.

        Args:
            name (TEncodable): The name of the field.
            type (FieldType): The type of the field.
            alias (Optional[TEncodable]): An alias for the field.
        """
        self.name = name
        self.type = type
        self.alias = alias

    @abstractmethod
    def to_args(self) -> List[TEncodable]:
        """
        Get the arguments representing the field.

        Returns:
            List[TEncodable]: A list of field arguments.
        """
        args = [self.name]
        if self.alias:
            args.extend([FtCreateKeywords.AS, self.alias])
        args.append(self.type.value)
        return args


class TextField(Field):
    """
    Field contains any blob of data.
    """

    def __init__(self, name: TEncodable, alias: Optional[TEncodable] = None):
        """
        Initialize a new TextField instance.

        Args:
            name (TEncodable): The name of the text field.
            alias (Optional[TEncodable]): An alias for the field.
        """
        super().__init__(name, FieldType.TEXT, alias)

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        return args


class TagField(Field):
    """
    Tag fields are similar to full-text fields, but they interpret the text as a simple list of
    tags delimited by a separator character.

    For `HASH fields, separator default is a comma `,`. For `JSON` fields, there is no
    default separator; you must declare one explicitly if needed.
    """

    def __init__(
        self,
        name: TEncodable,
        alias: Optional[TEncodable] = None,
        separator: Optional[TEncodable] = None,
        case_sensitive: bool = False,
    ):
        """
        Initialize a new TagField instance.

        Args:
            name (TEncodable): The name of the tag field.
            alias (Optional[TEncodable]): An alias for the field.
            separator (Optional[TEncodable]): Specify how text in the attribute is split into individual tags. Must be a
                single character.
            case_sensitive (bool): Preserve the original letter cases of tags. If set to False, characters are converted to
                lowercase by default.
        """
        super().__init__(name, FieldType.TAG, alias)
        self.separator = separator
        self.case_sensitive = case_sensitive

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        if self.separator:
            args.extend([FtCreateKeywords.SEPARATOR, self.separator])
        if self.case_sensitive:
            args.append(FtCreateKeywords.CASESENSITIVE)
        return args


class NumericField(Field):
    """
    Field contains a number.
    """

    def __init__(self, name: TEncodable, alias: Optional[TEncodable] = None):
        """
        Initialize a new NumericField instance.

        Args:
            name (TEncodable): The name of the numeric field.
            alias (Optional[TEncodable]): An alias for the field.
        """
        super().__init__(name, FieldType.NUMERIC, alias)

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        return args


class VectorFieldAttributes(ABC):
    """
    Abstract base class for defining vector field attributes to be used after the vector algorithm name.
    """

    @abstractmethod
    def __init__(
        self, dimensions: int, distance_metric: DistanceMetricType, type: VectorType
    ):
        """
        Initialize a new vector field attributes instance.

        Args:
            dimensions (int): Number of dimensions in the vector. Equivalent to `DIM` on the module API.
            distance_metric (DistanceMetricType): The distance metric used in vector type field. Can be one of
                `[L2 | IP | COSINE]`. Equivalent to `DISTANCE_METRIC` on the module API.
            type (VectorType): Vector type. The only supported type is `FLOAT32`. Equivalent to `TYPE` on the module API.
        """
        self.dimensions = dimensions
        self.distance_metric = distance_metric
        self.type = type

    @abstractmethod
    def to_args(self) -> List[TEncodable]:
        """
        Get the arguments to be used for the algorithm of the vector field.

        Returns:
            List[TEncodable]: A list of arguments.
        """
        args: List[TEncodable] = []
        if self.dimensions:
            args.extend([FtCreateKeywords.DIM, str(self.dimensions)])
        if self.distance_metric:
            args.extend([FtCreateKeywords.DISTANCE_METRIC, self.distance_metric.name])
        if self.type:
            args.extend([FtCreateKeywords.TYPE, self.type.name])
        return args


class VectorFieldAttributesFlat(VectorFieldAttributes):
    """
    Get the arguments to be used for the FLAT algorithm of the vector field.
    """

    def __init__(
        self,
        dimensions: int,
        distance_metric: DistanceMetricType,
        type: VectorType,
        initial_cap: Optional[int] = None,
    ):
        """
        Initialize a new flat vector field attributes instance.

        Args:
            dimensions (int): Number of dimensions in the vector. Equivalent to `DIM` on the module API.
            distance_metric (DistanceMetricType): The distance metric used in vector type field. Can be one of
                `[L2 | IP | COSINE]`. Equivalent to `DISTANCE_METRIC` on the module API.
            type (VectorType): Vector type. The only supported type is `FLOAT32`. Equivalent to `TYPE` on the module API.
            initial_cap (Optional[int]): Initial vector capacity in the index affecting memory allocation size of the index.
                Defaults to `1024`. Equivalent to `INITIAL_CAP` on the module API.
        """
        super().__init__(dimensions, distance_metric, type)
        self.initial_cap = initial_cap

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        if self.initial_cap:
            args.extend([FtCreateKeywords.INITIAL_CAP, str(self.initial_cap)])
        return args


class VectorFieldAttributesHnsw(VectorFieldAttributes):
    """
    Get the arguments to be used for the HNSW algorithm of the vector field.
    """

    def __init__(
        self,
        dimensions: int,
        distance_metric: DistanceMetricType,
        type: VectorType,
        initial_cap: Optional[int] = None,
        number_of_edges: Optional[int] = None,
        vectors_examined_on_construction: Optional[int] = None,
        vectors_examined_on_runtime: Optional[int] = None,
    ):
        """
        Initialize a new HNSW vector field attributes instance.

        Args:
            dimensions (int): Number of dimensions in the vector. Equivalent to `DIM` on the module API.
            distance_metric (DistanceMetricType): The distance metric used in vector type field. Can be one of
                `[L2 | IP | COSINE]`. Equivalent to `DISTANCE_METRIC` on the module API.
            type (VectorType): Vector type. The only supported type is `FLOAT32`. Equivalent to `TYPE` on the module API.
            initial_cap (Optional[int]): Initial vector capacity in the index affecting memory allocation size of the index.
                Defaults to `1024`. Equivalent to `INITIAL_CAP` on the module API.
            number_of_edges (Optional[int]): Number of maximum allowed outgoing edges for each node in the graph in each layer.
                Default is `16`, maximum is `512`. Equivalent to `M` on the module API.
            vectors_examined_on_construction (Optional[int]): Controls the number of vectors examined during index
                construction. Default value is `200`, Maximum value is `4096`. Equivalent to `EF_CONSTRUCTION` on the
                module API.
            vectors_examined_on_runtime (Optional[int]): Controls the number of vectors examined during query operations.
                Default value is `10`, Maximum value is `4096`. Equivalent to `EF_RUNTIME` on the module API.
        """
        super().__init__(dimensions, distance_metric, type)
        self.initial_cap = initial_cap
        self.number_of_edges = number_of_edges
        self.vectors_examined_on_construction = vectors_examined_on_construction
        self.vectors_examined_on_runtime = vectors_examined_on_runtime

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        if self.initial_cap:
            args.extend([FtCreateKeywords.INITIAL_CAP, str(self.initial_cap)])
        if self.number_of_edges:
            args.extend([FtCreateKeywords.M, str(self.number_of_edges)])
        if self.vectors_examined_on_construction:
            args.extend(
                [
                    FtCreateKeywords.EF_CONSTRUCTION,
                    str(self.vectors_examined_on_construction),
                ]
            )
        if self.vectors_examined_on_runtime:
            args.extend(
                [FtCreateKeywords.EF_RUNTIME, str(self.vectors_examined_on_runtime)]
            )
        return args


class VectorField(Field):
    """
    Class for defining vector field in a schema.
    """

    def __init__(
        self,
        name: TEncodable,
        algorithm: VectorAlgorithm,
        attributes: VectorFieldAttributes,
        alias: Optional[TEncodable] = None,
    ):
        """
        Initialize a new VectorField instance.

        Args:
            name (TEncodable): The name of the vector field.
            algorithm (VectorAlgorithm): The vector indexing algorithm.
            alias (Optional[TEncodable]): An alias for the field.
            attributes (VectorFieldAttributes): Additional attributes to be passed with the vector field after the
                algorithm name.
        """
        super().__init__(name, FieldType.VECTOR, alias)
        self.algorithm = algorithm
        self.attributes = attributes

    def to_args(self) -> List[TEncodable]:
        args = super().to_args()
        args.append(self.algorithm.value)
        if self.attributes:
            attribute_list = self.attributes.to_args()
            args.append(str(len(attribute_list)))
            args.extend(attribute_list)
        return args


class DataType(Enum):
    """
    Type of the index dataset.
    """

    HASH = "HASH"
    """
    Data stored in hashes, so field identifiers are field names within the hashes.
    """
    JSON = "JSON"
    """
    Data stored as a JSON document, so field identifiers are JSON Path expressions.
    """


class FtCreateOptions:
    """
    This class represents the input options to be used in the FT.CREATE command.
    All fields in this class are optional inputs for FT.CREATE.
    """

    def __init__(
        self,
        data_type: Optional[DataType] = None,
        prefixes: Optional[List[TEncodable]] = None,
    ):
        """
        Initialize the FT.CREATE optional fields.

        Args:
            data_type (Optional[DataType]): The index data type. If not defined a `HASH` index is created.
            prefixes (Optional[List[TEncodable]]): A list of prefixes of index definitions.
        """
        self.data_type = data_type
        self.prefixes = prefixes

    def to_args(self) -> List[TEncodable]:
        """
        Get the optional arguments for the FT.CREATE command.

        Returns:
            List[TEncodable]:
                List of FT.CREATE optional agruments.
        """
        args: List[TEncodable] = []
        if self.data_type:
            args.append(FtCreateKeywords.ON)
            args.append(self.data_type.value)
        if self.prefixes:
            args.append(FtCreateKeywords.PREFIX)
            args.append(str(len(self.prefixes)))
            for prefix in self.prefixes:
                args.append(prefix)
        return args



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import pytest
from glide.protobuf.command_request_pb2 import CommandRequest, RequestType
from glide.protobuf.response_pb2 import Response
from glide.protobuf_codec import PartialMessageException, ProtobufCodec


class TestProtobufCodec:
    def test_encode_decode_delimited(self):
        request = CommandRequest()
        request.callback_idx = 1
        request.single_command.request_type = RequestType.Set
        args = [
            "foo",
            "bar",
        ]

        bytes_args = [bytes(elem, encoding="utf8") for elem in args]
        request.single_command.args_array.args[:] = bytes_args
        b_arr = bytearray()
        ProtobufCodec.encode_delimited(b_arr, request)
        msg_len_varint = int(b_arr[0])
        assert msg_len_varint == 19
        assert len(b_arr) == msg_len_varint + 1
        offset = 0
        b_arr_view = memoryview(b_arr)
        parsed_request, new_offset = ProtobufCodec.decode_delimited(
            b_arr, b_arr_view, offset, CommandRequest
        )
        assert new_offset == len(b_arr)
        assert parsed_request.callback_idx == 1
        assert parsed_request.single_command.request_type == RequestType.Set
        assert parsed_request.single_command.args_array.args == bytes_args

    def test_decode_partial_message_fails(self):
        response = Response()
        response.callback_idx = 1
        b_arr = bytearray()
        ProtobufCodec.encode_delimited(b_arr, response)
        b_arr_view = memoryview(b_arr)
        with pytest.raises(PartialMessageException):
            ProtobufCodec.decode_delimited(b_arr[:1], b_arr_view[:1], 0, Response)

    def test_decode_partial_varint_fails(self):
        varint = ProtobufCodec._varint_bytes(10000000)
        b_arr_view = memoryview(varint)
        assert len(varint) > 1
        with pytest.raises(PartialMessageException):
            ProtobufCodec.decode_delimited(varint[:1], b_arr_view[:1], 0, Response)

    def test_encode_decode_varint(self):
        value = 10000000
        varint = ProtobufCodec._varint_bytes(value)
        decoded_varint, res_len = ProtobufCodec._decode_varint_32(varint, 0)
        assert res_len == len(varint)
        assert decoded_varint == value
